package fitbit

import (
	"context"
	"crypto/hmac"
	"crypto/sha1"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/url"
	"time"

	"github.com/octo/gfitsync/app"
	"golang.org/x/oauth2"
	oauth2fitbit "golang.org/x/oauth2/fitbit"
	"google.golang.org/appengine/log"
)

var oauth2Config = &oauth2.Config{
	ClientID:     "@FITBIT_CLIENT_ID@",
	ClientSecret: "@FITBIT_CLIENT_SECRET@",
	Endpoint:     oauth2fitbit.Endpoint,
	RedirectURL:  "https://fitbit-gfit-sync.appspot.com/fitbit/grant",
	Scopes:       []string{"activity"},
}

const csrfToken = "@CSRFTOKEN@"

func AuthURL() string {
	return oauth2Config.AuthCodeURL(csrfToken, oauth2.AccessTypeOffline)
}

func ParseToken(ctx context.Context, r *http.Request, u *app.User) error {
	if state := r.FormValue("state"); state != csrfToken {
		return fmt.Errorf("invalid state parameter: %q", state)
	}

	tok, err := oauth2Config.Exchange(ctx, r.FormValue("code"))
	if err != nil {
		return err
	}

	return u.SetToken(ctx, "Fitbit", tok)
}

func CheckSignature(ctx context.Context, payload []byte, rawSig string) bool {
	base64Sig, err := url.QueryUnescape(rawSig)
	if err != nil {
		log.Errorf(ctx, "QueryUnescape(%q) = %v", rawSig, err)
		return false
	}
	signatureGot, err := base64.StdEncoding.DecodeString(base64Sig)
	if err != nil {
		log.Errorf(ctx, "base64.StdEncoding.DecodeString(%q) = %v", base64Sig, err)
		return false
	}

	mac := hmac.New(sha1.New, []byte(oauth2Config.ClientSecret+"&"))
	mac.Write(payload)
	signatureWant := mac.Sum(nil)

	return hmac.Equal(signatureGot, signatureWant)
}

type Activity struct {
	ActivityID       int     `json:"activityId"`
	ActivityParentID int     `json:"activityParentId"`
	Calories         int     `json:"calories"`
	Description      string  `json:"description"`
	Distance         float64 `json:"distance"`
	Duration         int     `json:"duration"`
	HasStartTime     bool    `json:"hasStartTime"`
	IsFavorite       bool    `json:"isFavorite"`
	LogID            int     `json:"logId"`
	Name             string  `json:"name"`
	StartTime        string  `json:"startTime"`
	Steps            int     `json:"steps"`
}

type Distance struct {
	Activity string  `json:"activity"`
	Distance float64 `json:"distance"`
}

type ActivitySummary struct {
	Activities []Activity `json:"activities"`
	Goals      struct {
		CaloriesOut int     `json:"caloriesOut"`
		Distance    float64 `json:"distance"`
		Floors      int     `json:"floors"`
		Steps       int     `json:"steps"`
	} `json:"goals"`
	Summary struct {
		ActivityCalories     int        `json:"activityCalories"`
		CaloriesBMR          int        `json:"caloriesBMR"`
		CaloriesOut          int        `json:"caloriesOut"`
		MarginalCalories     int        `json:"marginalCalories"`
		Distances            []Distance `json:"distances"`
		Elevation            float64    `json:"elevation"`
		Floors               int        `json:"floors"`
		Steps                int        `json:"steps"`
		SedentaryMinutes     int        `json:"sedentaryMinutes"`
		LightlyActiveMinutes int        `json:"lightlyActiveMinutes"`
		FairlyActiveMinutes  int        `json:"fairlyActiveMinutes"`
		VeryActiveMinutes    int        `json:"veryActiveMinutes"`
	} `json:"summary"`
}

type Subscription struct {
	CollectionType string `json:"collectionType"`
	Date           string `json:"date"`
	OwnerID        string `json:"ownerId"`
	OwnerType      string `json:"ownerType"`
	SubscriptionID string `json:"subscriptionId"`
}

type Client struct {
	fitbitUserID string
	appUser      *app.User
	client       *http.Client
}

func NewClient(ctx context.Context, fitbitUserID string, u *app.User) (*Client, error) {
	tok, err := u.Token(ctx, "Fitbit")
	if err != nil {
		return nil, err
	}

	if fitbitUserID == "" {
		fitbitUserID = "-"
	}

	return &Client{
		fitbitUserID: fitbitUserID,
		appUser:      u,
		client:       oauth2Config.Client(ctx, tok),
	}, nil
}

func (c *Client) ActivitySummary(t time.Time) (*ActivitySummary, error) {
	url := fmt.Sprintf("https://api.fitbit.com/1/user/%s/activity/date/%s.json",
		c.fitbitUserID, t.Format("2006-01-02"))

	res, err := c.client.Get(url)
	if err != nil {
		return nil, err
	}
	defer res.Body.Close()

	var summary ActivitySummary
	if err := json.NewDecoder(res.Body).Decode(&summary); err != nil {
		return nil, err
	}

	return &summary, nil
}

func (c *Client) Subscribe(ctx context.Context, collection string) error {
	subscriberID, err := c.appUser.ID(ctx)
	if err != nil {
		return err
	}

	url := fmt.Sprintf("https://api.fitbit.com/1/user/%s/%s/apiSubscriptions/%s.json",
		c.fitbitUserID, collection, subscriberID)
	res, err := c.client.Post(url, "", nil)
	if err != nil {
		return err
	}
	defer res.Body.Close()

	if res.StatusCode >= 400 {
		data, _ := ioutil.ReadAll(res.Body)
		log.Errorf(ctx, "creating subscription failed: status %d %q", res.StatusCode, data)
		return fmt.Errorf("creating subscription failed")
	}

	return nil
}
